#:include 'common.fypp'
module check
  use constants, only: sp, dp
  use debug, only: error_t
  use matrix
  implicit none
  public

  contains

#:for id, dtype, prec in TYPES
#:set one = CONST(1.0, dtype, prec)
#:set zero = CONST(0.0, dtype, prec)
    !> Calculates the Frobenius norm of (A * V - U * M) where M is the matrix of singular values.
    !> matrix_left = U, matrix_right = V
    function svd_accuracy_${id}$(matrix_a, matrix_left, singular_values, matrix_right, err) result(accuracy)
      type(matrix_${id}$_t), intent(in) :: matrix_a, matrix_left, matrix_right
      real(${prec}$), intent(in) :: singular_values(:)
      type(error_t), intent(inout) :: err

      integer :: i
      real(dp) :: accuracy
      type(matrix_${id}$_t) :: matrix_singular_values, matrix_av, matrix_ul

      call matrix_av%alloc_like(matrix_a, err)
      if (err%has_error()) then
        accuracy = -1._dp
        return
      end if
      call matrix_ul%alloc_like(matrix_a, err)
      if (err%has_error()) then
        accuracy = -1._dp
        return
      end if
      call matrix_singular_values%alloc_like(matrix_right, err)
      if (err%has_error()) then
        accuracy = -1._dp
        return
      end if

      call matrix_av%mm(matrix_a, matrix_right)

      call matrix_singular_values%set(${zero}$, ${zero}$)
      do i = 1, size(singular_values)
        call matrix_singular_values%set_at(i, i, ${one}$ * singular_values(i))
      end do

      call matrix_ul%mm(matrix_left, matrix_singular_values)

      call matrix_singular_values%dealloc()

      call matrix_av%sub(matrix_ul)

      accuracy = real(matrix_av%norm() / maxval(singular_values), kind=dp)

      call matrix_ul%dealloc()
      call matrix_av%dealloc()
    end function

    !> Returns true if any of the singular values is less than zero
    function has_invalid_singular_value_${id}$(singular_values, err) result(invalid)
      real(${prec}$), intent(in) :: singular_values(:)
      type(error_t), intent(inout) :: err
      logical :: invalid

      invalid = any(singular_values < 0._${prec}$)
      if (invalid) then
        call err%add_error("Invalid singular value detected")
      end if
    end function
#:endfor

end module
